""" Command Line Utilities for Sermos Cloud Rho Model APIs
"""
import json
import logging
import click
from rho_ml import search_rho_model

logger = logging.getLogger(__name__)


@click.group()
def sermos_cloud():
    """ Sermos Cloud command group.
    """


@sermos_cloud.command()
@click.option('--model-name', required=False, default=None)
@click.option('--version-pattern', required=False, default=None)
@click.option('--access-key', required=False, default=None)
@click.option('--search-model-endpoint', required=False, default=None)
def search(access_key: str = None,
           model_name: str = None,
           version_pattern: str = None,
           search_model_endpoint: str = None):
    """ Search for available models against Sermos Admin's public API.

        Arguments::

            access-key (optional): Defaults to checking the environment for
                `SERMOS_ACCESS_KEY`. If not found, will exit.

            model-name (optional): Model name to search for. Accepts wildcard.
                e.g. MyModelName, MyModel* If none provided, will search
                against all model names.

            version-pattern (optional): Version to search for. Accepts wildcard.
                e.g. 1.1.1, 1.*.* If none provided, will search against all
                versions.

            search-url (optional): Defaults to primary Sermos search-models
                endpoint. Only modify this if there is a specific, known reason
                to do so.
    """
    resp = search_rho_model(model_name, version_pattern, access_key,
                            search_model_endpoint)
    click.echo(f'{json.dumps(resp, indent=2)}')



""" Base class definition for RhoModel

    RhoModel provides a standard interface for models, or any other artifact
    which should be stored and versioned.

    Definition of StoredModel container class for storage of RhoModel instances.

    A StoredModel needs to contain the data and logic to deserialize the
    underlying RhoModel.  In cases where the RhoModel instance requires
    specialized serialization logic (e.g. when it isn't pickleable), a
    StoredModel subclass should be created.
"""
import pickle
from typing import Any, Callable, Optional, Union, Dict

import attr

from rho_ml import Version, StoredModel


class ValidationFailedError(Exception):
    """ Error for when validation fails for either training or prediction
    input/output in a RhoModel. """
    pass


@attr.s(auto_attribs=True)
class RhoModel(object):
    """ The most basic wrapper for an ML or other model.  Specific models should
    be a subclass of Predictor, and implement whatever appropriate subset of
    abstract methods defined here.

    Args:
        version (Version): model version
        name (str, optional): model name, defaults to the `RhoModel` subclass
            name.  Used by serialization utils.

    Example::

        model = MyRhoModel(version=Version.from_string("0.0.1"))
        # name is 'MyRhoModel' because no name was passed
    """
    version: Union[Version, Dict[str, int]]
    name: Optional[str] = attr.ib(kw_only=True, default=None)

    def __attrs_post_init__(self):
        """ If attr.asdict is used to help serialize a model, this will
        convert the version back to a Version from dict.  Set the default name
         if self.name is None. """
        if isinstance(self.version, dict):
            self.version = Version(**self.version)
        if not self.name:
            self.name = str(self.__class__.__name__)

    @property
    def version_string(self) -> str:
        """ Convenience property to provide the version as a string instead
            of a Version object
        """
        return self.version.to_string()

    def validate_training_input(self, data: Any) -> None:
        """ Logic to validate data passed to self.train().

        Should raise a :class:`.ValidationError` if validation fails, otherwise
        do nothing.

        Args:
            data: Arbitrary training data

        Returns:
            None: this should raise if validation fails, otherwise do
            nothing.
        """
        raise NotImplementedError

    def validate_training_output(self, data: Any):
        """ Logic to validate data returned from self.train().

        Should raise a :class:`.ValidationError` if validation fails, otherwise
        do nothing.

        Args:
            data: Arbitrary training output

        Returns:
            None: this should raise if validation fails, otherwise do
            nothing. """
        raise NotImplementedError

    def validate_prediction_input(self, data: Any):
        """ Logic to validate data passed to self.predict().

        Should raise a :class:`.ValidationError` if validation fails, otherwise
        do nothing.

        Args:
            data: Arbitrary prediction input

        Returns:
            None: this should raise if validation fails, otherwise do
            nothing. """
        raise NotImplementedError

    def validate_prediction_output(self, data: Any):
        """ Logic to validate data returned from self.predict().

                Should raise a :class:`.ValidationError` if validation fails, otherwise
                do nothing.

                Args:
                    data: Arbitrary prediction input

                Returns:
                    None: this should raise if validation fails, otherwise do
                    nothing. """
        raise NotImplementedError

    def train_logic(self, training_data: Any, *args, **kwargs) -> Any:
        """ This method should be overridden with the appropriate logic to
        take training data, evaluation data, run training, and return
        relevant data (e.g. training and validation metrics for each epoch).

        Args:
            training_data: Arbitrary training data

        Returns:

        """
        raise NotImplementedError

    def train(self, training_data: Any, run_validation: bool, *args,
              **kwargs) -> Any:
        """ Default method for running training of a model on some arbitrary
        data.  Optionally validate inputs and outputs using
        self.validate_training_input and self.validate_training_output. """
        if run_validation:
            result = get_validated_result(
                model_method=self.train_logic,
                input_data=training_data,
                input_validator=self.validate_training_input,
                output_validator=self.validate_training_output,
                *args,
                **kwargs)
        else:
            result = self.train_logic(training_data=training_data,
                                      *args,
                                      **kwargs)
        return result

    def predict_logic(self, prediction_data: Any, *args, **kwargs) -> Any:
        """ This method should take a dict containing items with appropriate
        types for the model, and generate outputs with appropriate types."""
        raise NotImplementedError

    def predict(self, prediction_data: Any, run_validation: bool, *args,
                **kwargs) -> Any:
        """ Default method for running predictions on some arbitrary data,
        with optional validation of inputs and outputs. """
        if run_validation:
            result = get_validated_result(
                model_method=self.predict_logic,
                input_data=prediction_data,
                input_validator=self.validate_prediction_input,
                output_validator=self.validate_prediction_output,
                *args,
                **kwargs)
        else:
            result = self.predict_logic(prediction_data=prediction_data,
                                        *args,
                                        **kwargs)
        return result

    def serialize(self) -> bytes:
        """ Serialize the predictor object to a byte string.

        Note: this should be overridden if specialized serialization /
        deserialization is required. """
        return pickle.dumps(self, protocol=4)

    @classmethod
    def deserialize(cls, serialized: bytes):
        """ Instantiate a RhoModel object from a serialized byte string

        DEPRECATED: use build_stored_model instead """
        raise NotImplementedError

    def build_stored_model(self) -> StoredModel:
        """ Create a StoredModel object which can be used to properly
        reinstantiate the model later.

        Note: this *must* be implemented to use Sermos serialization utils
        """
        model_bytes = self.serialize()
        stored_model = StoredModel(model_bytes=model_bytes)
        return stored_model

    def save_to_disk(self, path_to_output: str):
        """ Logic to save a single Predictor object to disk."""
        serialized = self.serialize()
        with open(path_to_output, 'wb') as f:
            f.write(serialized)

    @classmethod
    def load_from_disk(cls, path_to_file: str):
        """ Logic to load the Predictor subclass from disk. """
        with open(path_to_file, 'rb') as f:
            loaded = cls.deserialize(f.read())
        return loaded


def get_validated_result(model_method: Callable, input_data: Any,
                         input_validator: Callable[[Any], None],
                         output_validator: Callable[[Any], None], *args,
                         **kwargs) -> Any:
    """ Helper function to generalize validation of training and
    prediction data """
    input_validator(input_data)
    output_data = model_method(input_data, *args, **kwargs)
    output_validator(output_data)
    return output_data



""" Rho ML primary public classes and methods.
"""
from .model_locator import Version, generate_model_locator, \
    validate_model_version, split_model_locator, \
    find_highest_compatible_version, find_matching_model_names
from .serialization import LocalModelStorage, PipelineStorageConfig, StoredModel
from .rho_model import RhoModel, ValidationFailedError
from .sermos_cloud import UnauthorizedError, ModelNotFoundError,\
    ModelFailedToStoreError, store_rho_model, load_rho_model, search_rho_model

__version__ = '0.9.0'



"""Classes responsible for serializing models in order to be stored.

This module defines a :class:`.StoredModel` which provides a standardized way to
interact with :class:`.RhoModel` instances.  An abstract base class
:class:`.PipelineStorageConfig` defines the behavior for storing models to
various data stores.
"""
import glob
import logging
import os
import tempfile
import pickle
from typing import Union

import attr

from rho_ml.model_locator import generate_model_locator, \
    find_highest_compatible_version

logger = logging.getLogger(__name__)


@attr.s(auto_attribs=True, frozen=True)
class StoredModel(object):
    """  Stores the bytes of a serialized model and is intended for storage.

    This is used as a generic format to store `RhoModel` instances.  The default
    behavior is to use pickle to serialize a given `RhoModel`, and then unpickle
    in `StoredModel`.  In cases where pickle is not appropriate for model
    serialization, a `StoredModel` child class should be created with the
    load_model method overridden.

    Attributes:
        model_bytes (bytes): The serialized byte string of some `RhoModel`
    """
    model_bytes: bytes = attr.ib(repr=False)

    def load_model(self) -> 'RhoModel':
        """ Use the instantiated StoredModel to load & return RhoModel instance.

            Note: this method should be overridden if the underlying RhoModel
            requires specialized deserialization logic.

            Returns:
                `RhoModel`: the deserialized model instance in storage

            Examples::
                my_stored_model = StoredModel.from_pickle(some_stored_bytes)
                my_rho_model = my_stored_model.load_model()
        """
        return pickle.loads(self.model_bytes)

    def to_pickle(self) -> bytes:
        """ Serialize the `StoredModel` instance (note: this is *not* for
        serializing the underlying RhoModel instance, which is serialized
        before instantiating the `StoredModel`.

        Returns:
            bytes: the pickled bytes of the `StoredModel` instance.

        Examples::
            serialized_stored_model = my_stored_model.to_pickle()
        """
        return pickle.dumps(self, protocol=4)

    @classmethod
    def from_pickle(cls, stored_bytes: bytes) -> 'StoredModel':
        """ Load a StoredModel from it's pickled bytes.

        This is the inverse of :meth:`~StoredModel.to_pickle`.

        Args:
            stored_bytes (bytes): The serialized `StoredModel`

        Returns:
            StoredModel: re-instantiated `StoredModel`
        """
        return pickle.loads(stored_bytes)


class PipelineStorageConfig(object):
    """ Base class for utility classes used to store RhoModel objects.

    TODO: Add example of when this is necessary and how it's used.
    """
    def store(self, model: 'RhoModel'):
        """ Store a RhoModel.

        Creates a StoredModel, and stores the pickle bytes of that StoredModel
        to the appropriate store.

        Args:
            model (`RhoModel`): The `RhoModel` instance to store

        Returns:
            NoneType
         """
        raise NotImplementedError

    def retrieve(self, *args, **kwargs) -> 'RhoModel':
        """ Pull the pickled bytes of the StoredModel from the data store and
        use that to instantiate the underlying RhoModel.
        """
        raise NotImplementedError

    def get_key_from_pattern(self, model_name: str,
                             version_pattern: str) -> str:
        """ Given some pattern of the form 'name-1.2.3', 'name-1.2.*', or
            'name-1.*', etc., return the matching key with the highest version.
            Returns None if nothing is found.

            Args:
                model_name (str): The name of the `RhoModel` search for
                version_pattern (str): Version (incl. wildcards) to search for

            Returns:
                 str: key of the highest matching artifact in the data store

            Examples::
                model_key = my_stored_model.get_key_from_pattern(
                'my_model_name', '0.*.*`)
                assert model_key == 'my_model_name_0.1.2'
        """
        raise NotImplementedError


@attr.s(auto_attribs=True)
class LocalModelStorage(PipelineStorageConfig):
    """ Class for caching `RhoModel` objects locally.

    Args:
        base_path (str): the directory to store, search, and/or retrieve models
            from.

    """
    base_path: str = attr.ib(default=None)

    def __attrs_post_init__(self):
        """ Provide default temp directory if no base_path set.
        """
        if not self.base_path:
            self.base_path = tempfile.gettempdir()

    def store(self, model: 'RhoModel'):
        """ Save pickled bytes of a RhoModel to local storage, in the
        `self.base_path` directory. """
        storage_key = generate_model_locator(
            model_name=model.name, model_version=model.version_string)
        storage_path = os.path.join(self.base_path, storage_key)
        stored_model = model.build_stored_model()
        store_bytes = stored_model.to_pickle()
        with open(storage_path, 'wb') as f:
            f.write(store_bytes)

    def retrieve(self, key: str) -> Union['RhoModel', None]:
        """ Attempt to retrieve model at a path that is stored locally. Return
            the loaded model if found, otherwise None.
        """
        storage_path = os.path.join(self.base_path, key)
        try:
            with open(storage_path, 'rb') as f:
                stored_model = StoredModel.from_pickle(stored_bytes=f.read())
                model = stored_model.load_model()
        except FileNotFoundError:
            return None
        return model

    def get_key_from_pattern(self, model_name: str,
                             version_pattern: str) -> Union[str, None]:
        """ Search `self.base_path` for an artifact matching the model name and
            pattern.
        """
        search_pattern = generate_model_locator(model_name=model_name,
                                                model_version=version_pattern)
        search_path = os.path.join(self.base_path, search_pattern)
        local_candidates = glob.glob(search_path)
        result_key = find_highest_compatible_version(
            search_version=version_pattern, search_list=local_candidates)
        return result_key



""" A "model locator" in Rho ML is the combination of the model name the model
    version, and a delimiter between them.

    This is important for storage and retrieval of models as they evolve over
    time. Using the default settings is highly recommended but each component is
    configurable.

    By default:

    #. Model names can be any alphanumeric character
    #. Delimeter is "_" (the underscore character)
    #. Model versions must adhere to [semver versioning](https://semver.org/)

    e.g.::

        MyModel_0.1.0

    Notes:

    #. The most commonly used member of this module is :class:`~.Version`,
       which is passed to any :class:`~rho_ml.rho_model.RhoModel` at
       instantiation.
    #. In most cases users won't need to interact directly with the other
       functions in this module.  One notable counterexample is if a custom
       :class:`~rho_ml.serialization.PipelineStorageConfig` is needed.

"""
import logging
import re
from typing import List, Tuple, Union, Optional
import attr

logger = logging.getLogger(__name__)

DEFAULT_MODEL_NAME_PATTERN = r"(.*)"
DEFAULT_DELIMITER_PATTERN = r"\_"
DEFAULT_VERSION_PATTERN = r"([\d\*]+\.[\d\*]+\.[\d\*]+.*)"


def generate_model_locator(
        model_name: str,
        model_version: str,
        delimiter: str = "_",
        model_name_pattern: Union[str, None] = DEFAULT_MODEL_NAME_PATTERN,
        delimiter_pattern: str = DEFAULT_DELIMITER_PATTERN,
        version_pattern: str = DEFAULT_VERSION_PATTERN) -> str:
    """ Combine the model name and version into name_version format, which
        by default uses "_" as the delimiter.

        Validates model name and version against the provided (or default)
        patterns.
    """
    # Validate model name against pattern
    if not re.compile(model_name_pattern).match(model_name):
        raise ValueError(f"{model_name} does not conform to "
                         f"model name pattern {model_name_pattern}")

    # Validate delimiter against pattern
    if not re.compile(delimiter_pattern).match(delimiter):
        raise ValueError(f"{delimiter} does not conform to "
                         f"delimiter pattern {delimiter_pattern}")

    # Validate model version against pattern
    if not re.compile(version_pattern).match(model_version):
        raise ValueError(f"{model_name} does not conform to "
                         f"model version pattern {version_pattern}")

    # Combine and return if all good
    return model_name + delimiter + model_version


def validate_model_version(
        model_version: str,
        version_pattern: str = DEFAULT_VERSION_PATTERN) -> str:
    """ Model version must conform to semver but can include wildcards.
    """
    pattern = re.compile(r"^" + version_pattern)
    if not pattern.match(model_version):
        raise ValueError("Invalid model_version format ({0}). Must conform "
                         "to semver with wildcards allowed. "
                         "e.g. 1.1.1, 1.*.*".format(model_version))
    return model_version


def split_model_locator(
        model_locator: str,
        model_name_pattern: Union[str, None] = DEFAULT_MODEL_NAME_PATTERN,
        delimiter_pattern: str = DEFAULT_DELIMITER_PATTERN,
        version_pattern: str = DEFAULT_VERSION_PATTERN,
        exclude_paths: bool = False
) -> Tuple[Union[str, None], Union[str, None]]:
    """ Given a full model locator including it's name and version (Model_0.1.0)
        return the model name and version as separate entities based on
        provided patterns (e.g. ('Model', '0.1.0')).

        Allows model_name_pattern to be None, in which case this expects the
        `model_locator` to instead be the version string along (e.g. 0.1.0)
        and will return (None, '0.1.0')

        exclude_paths allows you to pass the model_locator as a full path
        to the object. This is off by default because a model name can be
        "anything" according to the default model name pattern and this operates
        with the basic rule that the model_locator will exist after the final
        forward slash ('/')
    """
    version_search = re.compile(version_pattern)
    version_group = version_search.search(model_locator)
    if version_group is None:
        this_version = None
    else:
        this_version = version_group.group(1)

    if model_name_pattern is None:
        this_model_name = None
    else:
        model_name_search = re.compile(model_name_pattern + delimiter_pattern)
        model_name_group = model_name_search.search(model_locator)
        if model_name_group is None:
            this_model_name = None
        else:
            this_model_name = model_name_group.group(1)

    if exclude_paths:
        this_model_name = this_model_name.split('/')[-1]

    return this_model_name, this_version


def find_highest_compatible_version(
        search_version: str,
        search_list: List[str],
        model_name_pattern: Union[str, None] = DEFAULT_MODEL_NAME_PATTERN,
        delimiter_pattern: str = DEFAULT_DELIMITER_PATTERN,
        version_pattern: str = DEFAULT_VERSION_PATTERN) -> str:
    """ Return the model locator or version of the highest compatible version.

        If search_list is a list of model locators (['model_0.1.1']) it returns
        the full locator of the highest version that matches `search_version`.

        If search_list is a list of versions only (['0.1.1']), then it does
        the same search against `search_version` and returns just version number

        The default model locator pattern is a combination of the default
        model name pattern (any valid characters ending in "_" as the delimiter
        for the beginning of the version pattern) and the default version
        pattern (semver compatible 0.0.0*).

        That is::

            {{model name}}_{{model version}}

        e.g.::

            MyModel_0.1.0
            my_model_0.1.0
            prefix/mymodel_0.1.0

        If a different convention is followed, you can update the default
        model_name_pattern and version_pattern. Note: any custom search patterns
        must still adhere to some baked in assumptions

        #. The model name pattern must have a single capture group OR be
           None, which will result in this assuming it's only searching
           against a version pattern.
        #. The version pattern needs to follow semver, which is what the
           default pattern provides, so it's not recommended to change and
           also requires to have exactly 1 capture group.

        Args:
            search_version (string): The version to search. This can include
                wildcards. E.g. "1.0.0", "2.*.*", "1.1.*"
            search_list (list): List of full model locators with model name and
                version, e.g. `['model_name_0.0.0', 'model_name_0.1.0']`
                or a list of just version numbers, e.g. `['0.0.0', '0.1.0']`
            model_name_pattern (r string): Regex pattern describing any name
                that may be prepended to the search version. Pass as None
                to exclude a model name pattern (e.g. if your
                search_list is a list of only version nums
                ['0.1.1', '0.1.2']). NOTE: This pattern must currently consist
                of exactly one regex group.
            delimiter_pattern (r string): Regex pattern describing delimiter
                Not recommended to modify this pattern.
            version_pattern (r string): Regex pattern describing version
                numbers. NOT recommended to modify this pattern, as the logic
                assumes semver, which is default version pattern provided.
        Returns:
            string: string with major, minor, patch of highest
                compatible version found in available keys based on the
                search version. e.g. "0.1.0"
                e.g. if search version is "0.1.*" & available keys includes
                ['my_name_0.1.1', 'my_name_1.0.0', 'my_name_0.1.5']
                this will return my_name_0.1.5
    """
    search_version = validate_model_version(search_version)

    def _run_search(search_list, v_index, version_elements) -> Optional[str]:
        """ Loop through all keys to identify highest compatible.
            This is called sequentially for major, minor, patch.

            Args:
                search_list (list): List of full model locators with name and
                    version, e.g.
                        ['model_name_0.0.0', 'model_name_0.1.0']
                    or a list of just version numbers, e.g.
                        ['0.0.0', '0.1.0']
                v_index (int): The index of the model version to find. 0 is
                    major, 1 is minor, 2 is patch.
                version_elements (list): List of integers representing the
                    indices of version elements that have already been
                    found. The first element is always major, second is
                    minor, third is patch. e.g. [0, 1, 0] == "0.1.0"
            Returns:
                string: string with major, minor, patch of highest
                    compatible version found in available keys based on the
                    search version. e.g. "0.1.0"
        """
        highest_compatible_num = -1  # Haven't found anything yet
        for available_version in search_list:

            model_name, model_version = split_model_locator(
                available_version,
                model_name_pattern=model_name_pattern,
                delimiter_pattern=delimiter_pattern,
                version_pattern=version_pattern)

            if model_version is None:
                continue

            # Skip this available_version if it has a version element
            # > something we've already found. e.g. if we already found
            # major version 1 as the max, and this is major version 0,
            # then it doesn't matter what the value of this_num is because
            # it's already incompatible. This can only ever ben 3 elements
            # when following semver, so hardcode.
            incompatible_previous_version_num = False
            for i in range(len(version_elements)):
                if version_elements[i] is None \
                        or int(model_version.split(".")[i]) \
                        != version_elements[i]:
                    incompatible_previous_version_num = True

            if incompatible_previous_version_num:
                continue

            # We only compare one version num (major/minor/patch) at a time
            this_num = int(model_version.split(".")[v_index])

            # The 'search_num' can include wildcards, which means it
            # can be anything. If it's an int, then this_num must be
            # equal to the search_num.
            search_num = search_version.split(".")[v_index]
            if search_num == "*":
                if this_num > highest_compatible_num:
                    highest_compatible_num = this_num
            else:
                search_num = int(search_num)
                if this_num > highest_compatible_num \
                        and this_num == search_num:
                    highest_compatible_num = this_num

        if highest_compatible_num == -1:
            return None
        return highest_compatible_num

    version_elements = []
    for element_idx in range(3):
        version_elements.append(
            _run_search(search_list, element_idx, version_elements))

    if any(v is None for v in version_elements):
        msg = f"Unable to find compatible version with pattern {search_version}"
        if model_name_pattern is not None:
            msg += f" for model pattern {model_name_pattern}"
        msg += f" from available options: {search_list}"
        logger.warning(msg)
        return None

    highest_version = str(version_elements[0]) + "." \
                      + str(version_elements[1]) \
                      + "." + str(version_elements[2])

    # Return the whole key that has the discovered version.
    for key in search_list:
        model_name, model_version = split_model_locator(
            key,
            model_name_pattern=model_name_pattern,
            delimiter_pattern=delimiter_pattern,
            version_pattern=version_pattern)
        if model_version == highest_version:
            return key

    return None


def find_matching_model_names(
        search_pattern: str,
        available_model_locators: List[str],
        model_name_pattern: str = DEFAULT_MODEL_NAME_PATTERN,
        delimiter_pattern: str = DEFAULT_DELIMITER_PATTERN,
        version_pattern: str = DEFAULT_VERSION_PATTERN) -> List[str]:
    """ Return the model locators with model names that match the search_pattern

        By default, this expects that the model locators follow the convention

            {{model name}}_{{model version}}

        e.g.

            MyModel_0.1.0
            my_model_0.1.0
            prefix/mymodel_0.1.0

        If a different convention is followed, you can update the default
        model_name_pattern. The name pattern must have a single capture group.

        Args:
            search_pattern (string): The model name to look for. This can
                include wildcards. E.g. "MyModel", "*Model", "*"
            available_model_locators (list): List of full model locators with
                name and version, e.g.
                ['model_name_0.0.0', 'model_name_0.1.0', 'MyModel_0.1.0']
            model_name_pattern (r string): Regex pattern describing any name
                that may be prepended to the search version. Pass as None
                to return all keys. NOTE: This pattern must currently consist
                of exactly one regex group.
            delimiter_pattern (r string): Regex pattern describing delimiter
                Not recommended to modify this pattern.
        Returns:
            List[str]: List of model locators that match the search. e.g.
            if search pattern is "\*Model" & available keys includes
            ['my_name_0.1.1', 'my_name_1.0.0', 'MyModel_0.1.5'] this will
            return ['MyModel_0.1.5']
    """

    matched_locators = []
    for locator in available_model_locators:
        model_name, model_version = split_model_locator(
            locator, model_name_pattern, delimiter_pattern, version_pattern)
        if model_name is None:
            continue

        # We allow a star "*" as a wildcard indicator to be consistent with
        # version search.
        if '*' in search_pattern:
            wildcard_pattern = r''
            for char in search_pattern:
                if char == '*':
                    wildcard_pattern += r'.*'
                else:
                    wildcard_pattern += char

            # If the wildcard is NOT the first character, then prepend `^` to
            # search pattern. If wildcard is NOT the last character, then
            # append `$` to search pattern.
            if search_pattern[0] != '*':
                wildcard_pattern = r'^' + wildcard_pattern
            if search_pattern[-1] != '*':
                wildcard_pattern = wildcard_pattern + r'$'

            wildcard_pattern = re.compile(wildcard_pattern)
            if wildcard_pattern.search(model_name):
                matched_locators.append(locator)
        else:  # If no wildcard, do an exact match search
            if model_name == search_pattern:
                matched_locators.append(locator)

    return matched_locators


def convert_empty_str(val: Optional[str]) -> Optional[str]:
    """ Helper method to convert an empty string to None for consistency.
    """
    if not val:
        result = None
    else:
        result = val
    return result


@attr.s(auto_attribs=True, frozen=True)
class Version(object):
    """ Standard version class for RhoModel instances.

    Supports comparison (<, >, ==, etc.) """
    major: int
    minor: int
    patch: int
    label: Optional[str] = attr.ib(converter=convert_empty_str, default=None)

    separator_pattern = r'[\-\_\+\.]'  # Separate version from label
    label_pattern = r'((?P<separator>' \
                    + separator_pattern \
                    + r')(?P<label>.+))?'  # Add grp names to label & separator
    version_pattern = r'(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)' \
                      + label_pattern  # version groups & label groups combinee

    def __str__(self):
        return self.to_string()

    @property
    def label_no_separator(self):
        """ Convenience property to provide the label and exclude the separator

            e.g.
                label == '-alpha.1'
                label_no_separator == 'alpha.1'
        """
        if self.label:
            match = re.match(self.label_pattern, self.label)
            if match is not None:
                return match.group('label')
        return None

    @staticmethod
    def _get_to_increment(to_increment: str):
        """ Accepts a user defined to_increment value and ensures it's valid /
            gets default value.
        """
        if to_increment is None:
            to_increment = 'patch'
        allowable_options = ('major', 'minor', 'patch')
        if to_increment not in allowable_options:
            raise ValueError("Value of `to_increment` must be one of "
                             f"{allowable_options}")
        return to_increment

    @staticmethod
    def _get_new_version(version: 'Version', to_increment: str):
        """ Return the "new" major/minor/patch versions based on a provided
            'to_increment' value.
        """
        # Validate the to_increment val
        to_increment = version._get_to_increment(to_increment)

        if to_increment == 'major':
            major = version.major + 1
            minor = 0
            patch = 0
        elif to_increment == 'minor':
            major = version.major
            minor = version.minor + 1
            patch = 0
        elif to_increment == 'patch':
            major = version.major
            minor = version.minor
            patch = version.patch + 1
        else:
            raise ValueError(
                f"{to_increment} is not a valid version parameter "
                f"to bump!  Should be one of 'major', 'minor', "
                f"or 'patch'")
        return major, minor, patch

    @classmethod
    def increment_version(cls,
                          current_version: 'Version',
                          to_increment: Union[str, None] = None) -> 'Version':
        """ Given an old Version, increment the value of the version's
            `major`, `minor`, or `patch` value based on `to_increment`.

            If `to_increment` is None, defaults to `patch`

            This will always remove any label that may exist on the current
            version.
        """
        major, minor, patch = \
            cls._get_new_version(current_version, to_increment)
        return cls(major=major, minor=minor, patch=patch, label=None)

    @classmethod
    def from_string(cls, val: str):
        """ Create a Version object from a string of the form
            'major.minor.patch-label', where the label is optional.  If it is
            present, it must be separated from the rest of the string by either
            a hyphen (-), underscore (_), plus (+), or period (.).

            semver has specific guidelines around usage of "+_docs_build" and
            "-prerelease" but this level of granularity is not currently
            provided in Rho ML. It is major.minor.patch(optional additional of
            your choosing) where (optional additional) must start with one of the
            valid separators listed above.
        """
        match = re.match(cls.version_pattern, val)

        label = None
        if match.group('separator') is not None \
                and match.group('label') is not None:
            label = match.group('separator') + match.group('label')
        return cls(major=int(match.group('major')),
                   minor=int(match.group('minor')),
                   patch=int(match.group('patch')),
                   label=label)

    def to_string(self) -> str:
        """ Create a version string of the form
        major.min.patch[separator][label]
        where the separator and label are only present if both exist.
        """
        version_str = "{0}.{1}.{2}".format(self.major, self.minor, self.patch)
        if self.label:
            version_str += self.label
        return version_str
